/**
 * DiCloud - GENI project
 * 
 * (C) 2010 - University of Massachusetts Amherst
 * 
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package edu.umass.cs.geni.ec2.accounting;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;

import org.apache.log4j.Logger;

/**
 * This class does the accounting for all Amazon Web Services (AWS) operations.
 * The state is persisted in a Derby database with the following format:
 * 
 * <pre>
 * CREATE TABLE aws_activity (
 *  when      TIMESTAMP,     // when the operation occurred
 *  service   VARCHAR(20),   // AWS service: EC2, S3, EBS, Payment, ...
 *  operation VARCHAR(1000), // Operation performed
 *  cost      DOUBLE)        // Cost of the operation (negative if debit, positive if credit)
 * </pre>
 * 
 * @author <a href="mailto:cecchet@cs.umass.edu">Emmanuel Cecchet </a>
 * @version 1.0
 * 
 */
public class AWSAccounting
{

  private Logger            logger     = Logger.getLogger(AWSAccounting.class);

  /** Connection to the embedded Derby database */
  private Connection        conn;
  /** PreparedStatement used to log new entries in the aws_activity table */
  private PreparedStatement logPS;
  /** PreparedStatement used to retrieve the balance from the aws_activity table */
  private PreparedStatement balancePS;
  /** Minimum credit necessary to perform AWS operations */
  private double            minimumCredit;

  EC2MonitoringThread       ec2monitor = null;

  /**
   * Initialize the Derby database used for logging activity
   */
  public AWSAccounting(double minimumCredit)
  {
	this.minimumCredit = minimumCredit;
	try
	{
	  Class.forName("org.apache.derby.jdbc.EmbeddedDriver");
	  checkAndInitializeConnection();

	  createAwsActivityTable(conn);
	  initializeConnectionOrPreparedStatements();
	}
	catch (ClassNotFoundException e)
	{
	  logger.error("Could not find Derby in the classpath", e);
	}
	catch (SQLException e)
	{
	  logger.error("Failed to initialize Derby database", e);
	}
  }

  /**
   * Check if the connection is valid, and (re-)initialize it otherwise.
   * 
   * @throws SQLException
   *           if no connection could be obtained to the database
   */
  private void checkAndInitializeConnection() throws SQLException
  {
	if (conn != null)
	{ // Test the connection
	  try
	  {
		conn.createStatement().executeQuery("values 1");
		return; // success
	  }
	  catch (Exception e)
	  {
		// Bad connection renew it
	  }
	}
	conn = DriverManager.getConnection("jdbc:derby:geniawsdb;create=true");
  }

  /**
   * Initialize (or reinitialize after a failure) the Connection and the
   * PreparedStatements to the Derby database
   * 
   * @throws SQLException
   *           if an error occurs initializing
   */
  private void initializeConnectionOrPreparedStatements() throws SQLException
  {
	checkAndInitializeConnection();
	if (logPS != null)
	{ // Try to cleanup the resources
	  try
	  {
		logPS.close();
	  }
	  catch (Exception ignore)
	  {
	  }
	}
	logPS = conn.prepareStatement("INSERT INTO aws_activity VALUES (?,?,?,?)");
	if (balancePS != null)
	{ // Try to cleanup the resources
	  try
	  {
		balancePS.close();
	  }
	  catch (Exception ignore)
	  {
	  }
	}
	balancePS = conn.prepareStatement("SELECT SUM(cost) FROM aws_activity");
  }

  /**
   * Log the given activity in the database
   * 
   * @param activity
   *          The activity to log
   * @throws Exception
   *           if an error occurs logging the activity
   */
  public void logActivity(AWSActivity activity) throws Exception
  {
	try
	{
	  logPS.setTimestamp(1, new Timestamp(activity.getTime()));
	  logPS.setString(2, activity.getService());
	  logPS.setString(3, activity.getOperation());
	  logPS.setDouble(4, activity.getCost());
	  if (logPS.executeUpdate() != 1)
		logger.error("Unexpected number of rows update while logging activity "
		    + activity);

	  if (logger.isInfoEnabled())
	  {
		logger.info("Logging new activity @ " + activity.getTime());
		logger.info(" |-- Service: " + activity.getService());
		logger.info(" |-- Operation: " + activity.getOperation());
		logger.info(" '-- Cost: " + activity.getCost());
	  }
	}
	catch (SQLException e)
	{
	  logger.error("Failed to log activity" + activity, e);
	  try
	  {
		initializeConnectionOrPreparedStatements();
	  }
	  catch (SQLException e1)
	  {
		logger.error("Failed to re-initialize database connection", e1);
	  }
	  throw new Exception("Failed to log activity" + activity, e);
	}

	if (!hasCreditLeft() && (activity.getCost() < 0))
	{
	  logger.warn("AWS credit has expired, terminating all active resources.");
	  if (ec2monitor != null)
		ec2monitor.terminateActiveResources();
	}

  }

  /**
   * Log the given credit to the AWS account (when money is added to the
   * account)
   * 
   * @throws Exception
   *           if an error occurs logging the credit
   * 
   */
  public void logCredit(float credit) throws Exception
  {
	logActivity(new AWSActivity("Credit", "Payment", credit, System
	    .currentTimeMillis()));
  }

  /**
   * Retrieve the current balance of the account
   * 
   * @return the balance (positive if credit, negative if debit)
   * @throws Exception
   *           if an error occurs accessing the database
   */
  public double getBalance() throws Exception
  {
	try
	{
	  ResultSet rs = balancePS.executeQuery();
	  if (rs.next())
		return rs.getDouble(1);
	  else
		return 0; // Empty database
	}
	catch (SQLException e)
	{
	  logger.error("Failed to retrieve balance", e);
	  try
	  {
		initializeConnectionOrPreparedStatements();
	  }
	  catch (SQLException e1)
	  {
		logger.error("Failed to re-initialize database connection", e1);
	  }
	  throw new Exception("Failed to lretrieve balance", e);
	}
  }

  /**
   * @return the minimumCredit
   */
  public double getMinimumCredit()
  {
	return minimumCredit;
  }

  /**
   * @param minimumCredit
   *          the minimumCredit to set
   */
  public void setMinimumCredit(double minimumCredit)
  {
	this.minimumCredit = minimumCredit;
  }

  /**
   * Returns true if there is credit left for new AWS operations.
   * 
   * @return true if there is money left on the account!
   * @throws Exception
   *           if an error occurred getting the balance
   */
  public boolean hasCreditLeft() throws Exception
  {
	double balance = getBalance();
	if (logger.isInfoEnabled())
	  logger.info("Balance left: " + balance + " (minimum credit is "
		  + minimumCredit + ")");
	return balance >= minimumCredit;
  }

  private void createAwsActivityTable(Connection conn) throws SQLException
  {
	try
	{
	  conn
		  .createStatement()
		  .executeUpdate(
		      "CREATE TABLE aws_activity (when TIMESTAMP, service VARCHAR(20), operation VARCHAR(1000), cost DOUBLE)");
	}
	catch (SQLException e)
	{
	  if (!"X0Y32".equals(e.getSQLState())) // Table did not already exist
	  {
		logger.error("Failed to create aws_activity table ", e);
		throw e;
	  }
	}

  }

  /**
   * Start monitoring an EC2 instance for its disk, network and hourly usage.
   * 
   * @param instanceType
   *          EC2 instance type
   * @param region
   *          EC2 region
   * @param instanceId
   *          Id of the instance that has been launched
   */
  public void startEC2Monitoring(EC2Instance instance)
  {
	if (ec2monitor == null)
	{
	  ec2monitor = new EC2MonitoringThread(this);
	  ec2monitor.start();
	}

	ec2monitor.addInstance(instance);
  }

  /**
   * Empty the activity table
   * 
   * @throws SQLException
   *           if an error occurs
   */
  public void resetDatabase() throws SQLException
  {
	conn.createStatement().executeUpdate("DELETE FROM aws_activity");
  }

  /**
   * Shutdown all active resources and the database
   */
  public void shutdown()
  {
	if (ec2monitor != null)
	  ec2monitor.terminateActiveResources();
	try
	{
	  DriverManager.getConnection("jdbc:derby:geniawsdb;shutdown=true");
	}
	catch (SQLException e)
	{
	  logger.error("Failed to shutdown Derby", e);
	}
  }

}
