/**
 * DiCloud - GENI project
 * 
 * (C) 2010 - University of Massachusetts Amherst
 * 
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package edu.umass.cs.geni.ec2.accounting;

import java.util.ArrayList;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;

import edu.umass.cs.geni.ec2.NativeCommandExec;

/**
 * This thread monitors the network and disk activity of EC2 instances.
 * 
 * @author <a href="mailto:cecchet@cs.umass.edu">Emmanuel Cecchet </a>
 * @version 1.0
 * 
 */
public class EC2MonitoringThread extends Thread
{
  private Logger                 logger                             = Logger
	                                                                    .getLogger(AWSAccounting.class);

  private static final String    EC2_NETWORK_IN_MONITORING_COMMAND  = "mon-get-stats NetworkIn --namespace \"AWS/EC2\" --statistics \"Sum\"";
  private static final String    EC2_NETWORK_OUT_MONITORING_COMMAND = "mon-get-stats NetworkOut --namespace \"AWS/EC2\" --statistics \"Sum\"";
  private static final String    EC2_DISK_READ_MONITORING_COMMAND   = "mon-get-stats DiskReadOps --namespace \"AWS/EC2\" --statistics \"Sum\"";
  private static final String    EC2_DISK_WRITE_MONITORING_COMMAND  = "mon-get-stats DiskWriteOps --namespace \"AWS/EC2\" --statistics \"Sum\"";
  private static final String    EC2_PROBE_COMMAND                  = "ec2-describe-instances";
  private static final String    EC2_TERMINATE_COMMAND              = "ec2-terminate-instances";

  private AWSAccounting          awsAccounting;
  private ArrayList<EC2Instance> activeInstances                    = new ArrayList<EC2Instance>();
  private StringBuffer           netInTime                          = new StringBuffer(
	                                                                    19);
  private StringBuffer           netOutTime                         = new StringBuffer(
	                                                                    19);
  private StringBuffer           diskReadTime                       = new StringBuffer(
	                                                                    19);
  private StringBuffer           diskWriteTime                      = new StringBuffer(
	                                                                    19);

  private boolean                killed                             = false;

  /**
   * Create a new EC2 Monitoring thread
   * 
   * @param awsAccounting
   *          AWS accounting creating this thread
   */
  public EC2MonitoringThread(AWSAccounting awsAccounting)
  {
	super("EC2 Monitoring Thread");
	this.awsAccounting = awsAccounting;
  }

  /**
   * Monitor EC2 instances activity
   */
  public void run()
  {
	while (!killed)
	{
	  // Check for network activity
	  logger.info("Checking network activity of EC2 instances");
	  try
	  {
		double networkIn = runMonitoringCommand(
		    EC2_NETWORK_IN_MONITORING_COMMAND, netInTime);
		if (networkIn > 0)
		  awsAccounting.logActivity(new AWSActivity("EC2", "Network In",
			  networkIn / (1024 * 1024 * 1024)
			      * AWSActivityPricing.EC2_NETWORK_IN_PER_GB, System
			      .currentTimeMillis()));
	  }
	  catch (Exception e)
	  {
		logger.error("Monitoring command failed", e);
	  }
	  try
	  {
		double networkOut = runMonitoringCommand(
		    EC2_NETWORK_OUT_MONITORING_COMMAND, netOutTime);
		if (networkOut > 0)
		  awsAccounting.logActivity(new AWSActivity("EC2", "Network Out",
			  networkOut / (1024 * 1024 * 1024)
			      * AWSActivityPricing.EC2_NETWORK_OUT_PER_GB, System
			      .currentTimeMillis()));
	  }
	  catch (Exception e)
	  {
		logger.error("Monitoring command failed", e);
	  }

	  // Check for Disk activity
	  logger.info("Checking disk activity of EC2 instances");
	  try
	  {
		double diskRead = runMonitoringCommand(
		    EC2_DISK_READ_MONITORING_COMMAND, diskReadTime);
		if (diskRead > 0)
		  awsAccounting.logActivity(new AWSActivity("EC2", "Disk read",
			  diskRead / (double) 1000000
			      * AWSActivityPricing.EBS_IO_PER_MILLION, System
			      .currentTimeMillis()));
	  }
	  catch (Exception e)
	  {
		logger.error("Monitoring command failed", e);
	  }
	  try
	  {
		double diskWrite = runMonitoringCommand(
		    EC2_DISK_WRITE_MONITORING_COMMAND, diskWriteTime);
		if (diskWrite > 0)
		  awsAccounting.logActivity(new AWSActivity("EC2", "Disk write",
			  diskWrite / (double) 1000000
			      * AWSActivityPricing.EBS_IO_PER_MILLION, System
			      .currentTimeMillis()));
	  }
	  catch (Exception e)
	  {
		logger.error("Monitoring command failed", e);
	  }

	  // Check if instances have reached a new billed hour
	  logger.info("Checking EC2 instances hourly usage");
	  synchronized (this)
	  {
		long now = System.currentTimeMillis();
		for (EC2Instance instance : activeInstances)
		{
		  if (now - instance.getLastBilledTime() >= 3600 * 1000)
		  {
			// Check if the instance is still running
			String command = EC2_PROBE_COMMAND;
			// TODO: + " -K " + pk + " -C " + cert;
			if (instance.getRegion() != null)
			  command += " --region " + instance.getRegion();
			command += " " + instance.getInstanceId();

			NativeCommandExec exec = new NativeCommandExec();
			int status = -1;
			try
			{
			  status = exec.executeNativeCommand(command, null, null, 0, false);
			}
			catch (Exception e)
			{
			  logger.error("Failed to get instance " + instance.getInstanceId()
				  + " status", e);
			}

			debugCommandOutput(command, status, exec);

			boolean found = false;
			if (status == 0)
			{ // Success, check for information about the instance
			  for (String line : exec.getStdout())
			  {
				if (line.contains("INSTANCE"))
				{
				  StringTokenizer st = new StringTokenizer(line);
				  st.nextToken(); // INSTANCE
				  st.nextToken(); // instance id
				  st.nextToken(); // AMI id
				  st.nextToken(); // public name
				  st.nextToken(); // private name
				  String state = st.nextToken(); // state
				  if (state.equalsIgnoreCase("terminated"))
					activeInstances.remove(instance); // Instance terminated
				  // remove it
				  else
					found = true;
				}
			  }
			}

			if (found)
			  try
			  {
				awsAccounting
				    .logActivity(new AWSActivity("EC2", "Hourly usage",
				        AWSActivityPricing.getEC2InstancePrice(instance
				            .getInstanceType(), instance.getRegion()), now));
				instance.setLastBilledTime(now
				    - (now - instance.getLastBilledTime() - 3600000));
			  }
			  catch (Exception e)
			  {
				logger.error("Failed to log new hour usage for instance "
				    + instance.getInstanceId(), e);
			  }
		  }
		}
	  }

	  // Sleep for a minute (CloudWatch does not have a resolution finer than 1
	  // minute)
	  try
	  {
		Thread.sleep(60 * 1000);
	  }
	  catch (InterruptedException ignore)
	  {
	  }
	}
  }

  private double runMonitoringCommand(String command, StringBuffer startTime)
	  throws Exception
  {
	double result = 0;
	NativeCommandExec exec = new NativeCommandExec();

	try
	{
	  if (startTime.length() > 0)
		command += "  --start-time " + startTime;
	  int status = exec.executeNativeCommand(command, null, null, 0, false);
	  debugCommandOutput(command, status, exec);
	  if (status == 0)
	  { // Success, fetch the values
		for (String line : exec.getStdout())
		{
		  /**
		   * Sample output:
		   * 
		   * <pre>
		   * Time                 Samples  Sum       Unit
		   * 2010-07-13 03:20:00  2.0      0.0       Bytes
		   * </pre>
		   */
		  StringTokenizer st = new StringTokenizer(line);
		  String time = st.nextToken() + "T" + st.nextToken();
		  if (time.equals(startTime))
			continue; // Already accounted for => ignore
		  // Fill in the new time
		  startTime.delete(0, startTime.length());
		  startTime.append(time);
		  // Ignore samples
		  result += Double.valueOf(st.nextToken());
		}
		return result;
	  }
	  else
	  {
		String msg = "Monitoring command " + command + " was not succesful";
		logger.warn(msg);
		throw new Exception(msg);
	  }
	}
	catch (Exception e)
	{
	  logger.warn("Error while executing command " + command, e);
	  throw e;
	}

  }

  /**
   * Terminate all active EC2 instances and terminate this thread
   */
  public synchronized void terminateActiveResources()
  {
	for (EC2Instance instance : activeInstances)
	{
	  String command = EC2_TERMINATE_COMMAND;
	  // TODO: + " -K " + pk + " -C " + cert;
	  if (instance.getRegion() != null)
		command += " --region " + instance.getRegion();
	  command += " " + instance.getInstanceId();

	  if (logger.isInfoEnabled())
		logger.info("Terminating instance " + instance.getInstanceId());

	  NativeCommandExec exec = new NativeCommandExec();
	  try
	  {
		int status = exec.executeNativeCommand(command, null, null, 0, false);
		debugCommandOutput(command, status, exec);
	  }
	  catch (Exception e)
	  {
		logger.warn("Error while terminating instance "
		    + instance.getInstanceId(), e);
	  }
	}
	killed = true;
  }

  /**
   * Add an EC2 instance to monitor
   * 
   * @param instance
   *          the instance to monitor
   */
  public synchronized void addInstance(EC2Instance instance)
  {
	activeInstances.add(instance);
  }

  /**
   * Debug the output of a command being executed
   * 
   * @param commandName
   *          name of the command executed
   * @param status
   *          status returned by the command execution
   * @param exec
   *          execution object
   */
  private void debugCommandOutput(String commandName, int status,
	  NativeCommandExec exec)
  {
	if (logger.isDebugEnabled())
	{
	  logger.debug(commandName + " status: " + status);
	  logger.debug("stdout:" + exec.getStdout());
	  logger.debug("stderr:" + exec.getStderr());
	}

	if (status != 0)
	{
	  System.err.println("Command: " + commandName);
	  System.out.println("stdout:" + exec.getStdout());
	  System.out.println("stderr:" + exec.getStderr());
	}
  }

}
