/**
 * DiCloud - GENI project
 * 
 * (C) University of Massachusetts Amherst
 * 
 * Original contributor: Emmanuel Cecchet
 */
package edu.umass.cs.geni.ec2.handlers;

import java.util.MissingResourceException;
import java.util.Properties;
import java.util.StringTokenizer;

import orca.shirako.plugins.config.Config;
import orca.util.PropList;
import edu.umass.cs.geni.ec2.AmazonProperties;
import edu.umass.cs.geni.ec2.NativeCommandExec;

/**
 * This class defines an EBS handler to manage Amazon Elastic Block Storage
 * (EBS) volumes
 * 
 * @author <a href="mailto:cecchet@cs.umass.edu">Emmanuel Cecchet </a>
 * @version 1.0
 * 
 */
public class EBSHandler extends Config
{
  private static final String CREATE_EBS_VOLUME         = "ec2-create-volume";
  private static final String EBS_DELETE_VOLUME_COMMAND = "ec2-delete-volume";
  private static final String EBS_PROBE_COMMAND         = "ec2-describe-volumes";
  private AWSPlugin           awsPlugin;

  /**
   * Constructor just invoking super
   * 
   * @param plugin
   *          a Shirako plugin
   * 
   */
  public EBSHandler(AWSPlugin plugin)
  {
	super(plugin);
	this.awsPlugin = plugin;
  }

  /**
   * Create a new EBS volume by invoking ec2-create-volume.
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EBS_SIZE: AMI id to use to create instances
   *  - AmazonProperties.AVAILABILITY_ZONE: availability zone where to create the instance
   * Optional properties: 
   *  - AmazonProperties.REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: command output (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EBS_VOLUME_ID: Id of the EBS volume
   * AmazonProperties.EBS_CREATION_TIME: Creation time of the volume
   * AmazonProperties.REGION: region where the volume has been created
   * AmazonProperties.EBS_STATUS: status of the volume
   * AmazonProperties.EBS_CREATION_TIME: volume creation time
   * </pre>
   */
  @Override
  public void join(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);
	String availZone = p.getProperty(AmazonProperties.EC2_AVAILABILITY_ZONE);
	String size = p.getProperty(AmazonProperties.EBS_SIZE_IN_GB);

	// Sanity checks
	if (size == null)
	  throw new MissingResourceException("Missing EBS volume size in join",
		  this.getClass().getName(), AmazonProperties.EBS_SIZE_IN_GB);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in join",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in join", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);
	if (availZone == null)
	  throw new MissingResourceException("Missing availability zone in join",
		  this.getClass().getName(), AmazonProperties.EC2_AVAILABILITY_ZONE);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, TargetJoin);

	// Now create the instances
	if (logger.isInfoEnabled())
	  logger.info("Creating a " + size + " GB EBS Volume");

	String command = CREATE_EBS_VOLUME + " -K " + pk + " -C " + cert + " -s "
	    + size;
	if (region != null)
	  command += " --region " + region;
	if (availZone != null)
	  command += " -z" + availZone;

	// Let's execute the command
	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	if (status == 0)
	{ // Success, check for information about the instance
	  parseOutputAndSetProperties(result, exec);
	}
	else
	{ // Command failed
	  prepareResultError(CREATE_EBS_VOLUME, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Delete an EBS volume by invoking ec2-delete-volume
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EBS_VOLUME_ID: EBS volume id to delete
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: command output (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * </pre>
   */
  @Override
  public void leave(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String volumeId = p.getProperty(AmazonProperties.EBS_VOLUME_ID);
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);

	// Sanity checks
	if (volumeId == null)
	  throw new MissingResourceException("Missing volume id in leave", this
		  .getClass().getName(), AmazonProperties.EBS_VOLUME_ID);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in leave",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in leave", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);

	if (logger.isInfoEnabled())
	  logger.info("Deleting volume " + volumeId);

	// Let's execute the command
	String command = EBS_DELETE_VOLUME_COMMAND + " -K " + pk + " -C " + cert;
	if (region != null)
	  command += " --region " + region;
	command += " " + volumeId;

	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, TargetLeave);

	if (status == 0)
	{ // Success, return stdout
	  result.setProperty(Config.PropertyTargetResultCode, Integer
		  .toString(ResultCodeOK));
	  result.setProperty(Config.PropertyTargetResultCodeMessage, exec
		  .getStdout().toString());
	}
	else
	{ // Command failed
	  prepareResultError(EBS_DELETE_VOLUME_COMMAND, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Invoke ec2-describe-volumes to probe the EBS volume. The expected
   * properties are:
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.EC2_AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.EC2_AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EBS_VOLUME_ID: EBS volume id to probe
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: AMI status (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EBS_VOLUME_ID: Id of the EBS volume
   * AmazonProperties.EBS_CREATION_TIME: Creation time of the volume
   * AmazonProperties.REGION: region where the volume has been created
   * AmazonProperties.EBS_STATUS: status of the volume
   * AmazonProperties.EBS_CREATION_TIME: volume creation time
   * </pre>
   */
  public void probe(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String volumeId = p.getProperty(AmazonProperties.EBS_VOLUME_ID);
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);

	// Sanity checks
	if (volumeId == null)
	  throw new MissingResourceException("Missing volume id in probe", this
		  .getClass().getName(), AmazonProperties.EBS_VOLUME_ID);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in probe",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in probe", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);

	if (logger.isInfoEnabled())
	  logger.info("Probing volume " + volumeId);

	// Let's execute the command
	String command = EBS_PROBE_COMMAND + " -K " + pk + " -C " + cert;
	if (region != null)
	  command += " --region " + region;
	command += " " + volumeId;

	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, AWSPlugin.TargetProbe);

	if (status == 0)
	{ // Success, check for information about the instance
	  parseOutputAndSetProperties(result, exec);
	}
	else
	{ // Command failed
	  prepareResultError(EBS_PROBE_COMMAND, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Parse the output of the EBS create or describe commands
   * 
   * @param result
   * @param exec
   */
  private void parseOutputAndSetProperties(Properties result,
	  NativeCommandExec exec)
  {
	StringTokenizer st = new StringTokenizer(exec.getStdout().toString());
	int fieldNb = 0;
	while (st.hasMoreTokens())
	{
	  fieldNb++;
	  String field = st.nextToken();
	  // System.out.println("Field[" + fieldNb + "]=" + field);
	  switch (fieldNb)
	  {
	  // 1: VOLUME
	  case 2:
		result.setProperty(AmazonProperties.EBS_VOLUME_ID, field);
		break;
	  // 3: index
	  case 4:
		result.setProperty(AmazonProperties.EC2_REGION, field);
		break;
	  case 5:
		result.setProperty(AmazonProperties.EBS_STATUS, field);
		break;
	  case 6:
		result.setProperty(AmazonProperties.EBS_CREATION_TIME, field);
		break;
	  default:
		break;
	  }
	}

	result.setProperty(Config.PropertyTargetResultCode, Integer
	    .toString(ResultCodeOK));
	result.setProperty(Config.PropertyTargetResultCodeMessage, exec.getStdout()
	    .toString());
  }

  /**
   * Completion of the commands and return of the results
   * 
   * @param token
   * @param p
   * @param result
   */
  private void returnResult(Object token, Properties p, Properties result)
  {
	// pass the configuration sequence number back
	PropList.setProperty(result, Config.PropertyActionSequenceNumber, p
	    .getProperty(Config.PropertyActionSequenceNumber));
	plugin.configurationComplete(token, result);
  }

  /**
   * Add the proper error code and error message to be returned
   * 
   * @param commandName
   *          name of the command executed
   * @param exec
   *          execution object
   * @param result
   *          properties that are returned
   */
  private void prepareResultError(String command, NativeCommandExec exec,
	  Properties result)
  {
	result.setProperty(Config.PropertyTargetResultCode, Integer
	    .toString(ResultCodeException));
	result.setProperty(PropertyExceptionMessage, command + " execution failed");
	result.setProperty(PropertyExceptionStack, exec.getStderr().toString());
  }

  /**
   * Debug the output of a command being executed
   * 
   * @param commandName
   *          name of the command executed
   * @param status
   *          status returned by the command execution
   * @param exec
   *          execution object
   */
  private void debugCommandOutput(String commandName, int status,
	  NativeCommandExec exec)
  {
	if (logger.isDebugEnabled())
	{
	  logger.debug(commandName + " status: " + status);
	  logger.debug("stdout:" + exec.getStdout());
	  logger.debug("stderr:" + exec.getStderr());
	}

	System.err.println("Command: " + commandName);
	System.out.println("stdout:" + exec.getStdout());
	System.out.println("stderr:" + exec.getStderr());
  }

}
