/**
 * DiCloud - GENI project
 * 
 * (C) University of Massachusetts Amherst
 * 
 * Original contributor: Emmanuel Cecchet
 */
package edu.umass.cs.geni.ec2.handlers;

import java.io.IOException;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.StringTokenizer;

import orca.shirako.plugins.config.Config;
import orca.util.PropList;

import org.apache.log4j.Logger;

import edu.umass.cs.geni.ec2.AmazonProperties;
import edu.umass.cs.geni.ec2.NativeCommandExec;
import edu.umass.cs.geni.ec2.accounting.AWSActivity;
import edu.umass.cs.geni.ec2.accounting.AWSActivityPricing;
import edu.umass.cs.geni.ec2.accounting.EC2Instance;

/**
 * This class defines an EC2 handler to manage EC2 instances (servers)
 * 
 * @author <a href="mailto:cecchet@cs.umass.edu">Emmanuel Cecchet </a>
 * @version 1.0
 * 
 */
public class EC2Handler extends Config
{
  private static final String EC2_DELETE_KEYPAIR_COMMAND     = "ec2-delete-keypair";
  private static final String EC2_ADD_KEYPAIR_COMMAND        = "ec2-add-keypair";
  private static final String EC2_START_INSTANCE_COMMAND     = "ec2-run-instances";
  private static final String EC2_TERMINATE_INSTANCE_COMMAND = "ec2-terminate-instances";
  private static final String EC2_PROBE_COMMAND              = "ec2-describe-instances";
  private static final String KEY_PAIR_PREFIX                = "geni_dicloud_";
  private AWSPlugin           awsPlugin;
  private Logger              keyPairLogger                  = Logger
	                                                             .getLogger("edu.umass.cs.geni.ec2.handlers.keypair");

  /**
   * Constructor just invoking super
   * 
   * @param plugin
   *          a Shirako plugin
   * 
   */
  public EC2Handler(AWSPlugin plugin)
  {
	super(plugin);
	this.awsPlugin = plugin;
  }

  /**
   * Create a new EC2 instance by invoking ec2-run-instances.
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EC2_AMI_ID: AMI id to use to create instances
   *  - AmazonProperties.EC2_INSTANCE_TYPE: type of the instance to create
   *  - AmazonProperties.EC2_INSTANCE_NUMBER: number of instances to create
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   *  - AmazonProperties.EC2_AVAILABILITY_ZONE: availability zone where to create the instance
   *  - AmazonProperties.EC2_KEY_PAIR_NAME: registered key pair name to use (one will be generated 
   *  automatically and returned in AmazonProperties.EC2_KEY_PAIR if none is provided)
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: command output (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EC2_INSTANCE_ID: Instance id
   * AmazonProperties.EC2_KEY_PAIR_NAME: Name of the key pair that has been registered
   * AmazonProperties.EC2_KEY_PAIR: Key pair created for this instance
   * </pre>
   */
  @Override
  public void join(Object token, Properties p) throws Exception
  {
	if (!awsPlugin.getAwsAccouting().hasCreditLeft())
	  throw new Exception("No AWS credit left to allocate resources");

	// Prepare the command to be executed
	String ami = p.getProperty(AmazonProperties.EC2_AMI_ID);
	String instance_nb = p.getProperty(AmazonProperties.EC2_INSTANCE_NUMBER);
	String instance_type = p.getProperty(AmazonProperties.EC2_INSTANCE_TYPE);
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);
	String availZone = p.getProperty(AmazonProperties.EC2_AVAILABILITY_ZONE);
	String keyName = p.getProperty(AmazonProperties.EC2_KEY_PAIR_NAME);

	// Sanity checks
	if (ami == null)
	  throw new MissingResourceException("Missing instance id in join", this
		  .getClass().getName(), AmazonProperties.EC2_AMI_ID);
	if (instance_nb == null)
	  throw new MissingResourceException(
		  "Missing number of instances to create in join", this.getClass()
		      .getName(), AmazonProperties.EC2_INSTANCE_NUMBER);
	else if (Integer.valueOf(instance_nb) != 1)
	  throw new Exception(
		  "Number of EC2 instances to create currently limited to 1 in join");
	if (instance_type == null)
	  throw new MissingResourceException(
		  "Missing instance type to create in join", this.getClass().getName(),
		  AmazonProperties.EC2_INSTANCE_TYPE);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in join",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in join", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, TargetJoin);

	if (keyName == null)
	{// First create a keypair
	  keyName = KEY_PAIR_PREFIX + System.currentTimeMillis();
	  result.setProperty(AmazonProperties.EC2_KEY_PAIR_NAME, keyName);
	  createKeyPair(pk, cert, region, result, keyName);
	}

	// Now create the instances
	if (logger.isInfoEnabled())
	  logger.info("Creating " + instance_nb + " instance(s) from AMI " + ami);

	String command = EC2_START_INSTANCE_COMMAND + " -K " + pk + " -C " + cert;
	if (region != null)
	  command += " --region " + region;
	// -m for monitoring
	command += " " + ami + " -m -n " + instance_nb + " -k " + keyName + " -t "
	    + instance_type;
	if (availZone != null)
	  command += " -z " + availZone;

	// Let's execute the command
	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	if (status == 0)
	{ // Success, check for information about the instance
	  String instanceId = null;
	  for (String line : exec.getStdout())
	  {
		if (line.contains("INSTANCE"))
		{ // Extract the instance id
		  StringTokenizer st = new StringTokenizer(line);
		  st.nextToken(); // INSTANCE
		  instanceId = st.nextToken();
		  result.setProperty(AmazonProperties.EC2_INSTANCE_ID, instanceId);
		  if (logger.isDebugEnabled())
			logger.debug("instance id: '" + instanceId + "'");
		  System.out.println("instance id: '" + instanceId + "'");
		}
	  }

	  // Log the first hour of these instances
	  awsPlugin.getAwsAccouting().logActivity(
		  new AWSActivity("EC2", command, AWSActivityPricing
		      .getEC2InstancePrice(instance_type, region)
		      * Double.valueOf(instance_nb), System.currentTimeMillis()));

	  // Start the EC2 monitoring thread
	  awsPlugin.getAwsAccouting().startEC2Monitoring(
		  new EC2Instance(instance_type, region, instanceId));

	  result.setProperty(Config.PropertyTargetResultCode, Integer
		  .toString(ResultCodeOK));
	  result.setProperty(Config.PropertyTargetResultCodeMessage, exec
		  .getStdout().toString());
	}
	else
	{ // Command failed
	  prepareResultError(EC2_START_INSTANCE_COMMAND, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Stop an EC2 instance by invoking ec2-terminate-instances
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EC2_INSTANCE_ID: AMI instance id to probe
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   *  - AmazonProperties.EC2_KEY_PAIR_NAME: Key pair name used for this instance
   *  (if the keypair is a temporary keypair starting with KEY_PAIR_PREFIX, it is deleted)
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: command output (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * </pre>
   */
  @Override
  public void leave(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String ami = p.getProperty(AmazonProperties.EC2_INSTANCE_ID);
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);
	String keyName = p.getProperty(AmazonProperties.EC2_KEY_PAIR_NAME);

	// Sanity checks
	if (ami == null)
	  throw new MissingResourceException("Missing instance id in leave", this
		  .getClass().getName(), AmazonProperties.EC2_AMI_ID);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in leave",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in leave", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);

	if (logger.isInfoEnabled())
	  logger.info("Terminating instance " + ami);

	// Let's execute the command
	String command = EC2_TERMINATE_INSTANCE_COMMAND + " -K " + pk + " -C "
	    + cert;
	if (region != null)
	  command += " --region " + region;
	command += " " + ami;

	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, TargetLeave);

	if (status == 0)
	{ // Success, return stdout
	  result.setProperty(Config.PropertyTargetResultCode, Integer
		  .toString(ResultCodeOK));
	  result.setProperty(Config.PropertyTargetResultCodeMessage, exec
		  .getStdout().toString());
	}
	else
	{ // Command failed
	  prepareResultError(EC2_TERMINATE_INSTANCE_COMMAND, exec, result);
	}

	// Delete the temporary keypair
	if ((keyName != null) && (keyName.startsWith(KEY_PAIR_PREFIX)))
	  deleteKeyPair(pk, cert, region, keyName, result);

	returnResult(token, p, result);
  }

  /**
   * Invoke ec2-describe-instances to probe the instance. The expected
   * properties are:
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.EC2_AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.EC2_AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EC2_INSTANCE_ID: AMI instance id to probe
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: AMI status (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EC2_INSTANCE_ID: Instance id
   * AmazonProperties.EC2_AMI_ID: AMI id
   * AmazonProperties.EC2_PUBLIC_NAME: Public name
   * AmazonProperties.EC2_PRIVATE_NAME: Private name
   * AmazonProperties.EC2_PUBLIC_IP: Public IP
   * AmazonProperties.EC2_PRIVATE_IP: Private IP
   * AmazonProperties.EC2_INSTANCE_STATE: Instance state
   * AmazonProperties.EC2_INSTANCE_TYPE: Instance type
   * AmazonProperties.EC2_KEY_PAIR_NAME: Name of the key pair used
   * AmazonProperties.EC2_START_TIME: Start time
   * AmazonProperties.EC2_AVAILABILITY_ZONE: Availability zone
   * </pre>
   */
  public void probe(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String ami = p.getProperty(AmazonProperties.EC2_INSTANCE_ID);
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);

	// Sanity checks
	if (ami == null)
	  throw new MissingResourceException("Missing instance id in probe", this
		  .getClass().getName(), AmazonProperties.EC2_AMI_ID);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in probe",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in probe", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);

	if (logger.isInfoEnabled())
	  logger.info("Probing instance " + ami);

	// Let's execute the command
	String command = EC2_PROBE_COMMAND + " -K " + pk + " -C " + cert;
	if (region != null)
	  command += " --region " + region;

	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, AWSPlugin.TargetProbe);

	if (status == 0)
	{ // Success, check for information about the instance
	  boolean found = false;
	  for (String line : exec.getStdout())
	  {
		if (line.contains("INSTANCE") && line.contains(ami))
		{
		  result.setProperty(Config.PropertyTargetResultCode, Integer
			  .toString(ResultCodeOK));
		  result.setProperty(Config.PropertyTargetResultCodeMessage, line);
		  StringTokenizer st = new StringTokenizer(line);
		  int fieldNb = 0;
		  String fields[] = new String[st.countTokens()];
		  while (st.hasMoreTokens())
		  {
			fields[fieldNb++] = st.nextToken();
		  }
		  // 0: INSTANCE
		  result.setProperty(AmazonProperties.EC2_INSTANCE_ID, fields[1]);
		  result.setProperty(AmazonProperties.EC2_AMI_ID, fields[2]);
		  int indexShift = 0;
		  // There is no name of IP information for pending or terminated
		  // instances
		  if (line.contains("pending") || line.contains("terminated"))
			indexShift = -2;

		  // AMI status: INSTANCE i-6f9e4505 ami-78f21911 pending
		  // geni_dicloud_1279642629716 0 m1.small 2010-07-20T16:17:26+0000
		  // us-east-1a aki-a71cf9ce ari-a51cf9cc
		  // monitoring-enabledinstance-store

		  // AMI status: INSTANCE i-6f9e4505 ami-78f21911 terminated
		  // geni_dicloud_12796426297160m1.small 2010-07-20T16:17:26+0000
		  // us-east-1a aki-a71cf9ce ari-a51cf9cc monitoring-enabled
		  // instance-store

		  if (indexShift == 0)
		  {
			result.setProperty(AmazonProperties.EC2_PUBLIC_NAME,
			    fields[3 + indexShift]);
			result.setProperty(AmazonProperties.EC2_PRIVATE_NAME,
			    fields[4 + indexShift]);
		  }
		  result.setProperty(AmazonProperties.EC2_INSTANCE_STATE,
			  fields[5 + indexShift]);
		  result.setProperty(AmazonProperties.EC2_KEY_PAIR_NAME,
			  fields[6 + indexShift]);
		  // 7: AMI launch index
		  result.setProperty(AmazonProperties.EC2_INSTANCE_TYPE,
			  fields[8 + indexShift]);
		  result.setProperty(AmazonProperties.EC2_START_TIME,
			  fields[9 + indexShift]);
		  result.setProperty(AmazonProperties.EC2_AVAILABILITY_ZONE,
			  fields[10 + indexShift]);
		  // 11: aki
		  // 12: ari
		  // 13: monitoring info
		  if (fields.length >= 16 + indexShift)
		  {
			result.setProperty(AmazonProperties.EC2_PUBLIC_IP,
			    fields[14 + indexShift]);
			result.setProperty(AmazonProperties.EC2_PRIVATE_IP,
			    fields[15 + indexShift]);
		  }
		  found = true;
		  break;
		}
	  }
	  if (!found)
	  {
		result.setProperty(Config.PropertyTargetResultCode, Integer
		    .toString(ResultCodeException));
		result.setProperty(PropertyExceptionMessage,
		    "ec2-describe-instances did not report information about AMI id "
		        + ami);
	  }
	}
	else
	{ // Command failed
	  prepareResultError(EC2_PROBE_COMMAND, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * @param pk
   * @param cert
   * @param region
   * @param result
   * @param keyName
   * @throws IOException
   * @throws InterruptedException
   */
  private void createKeyPair(String pk, String cert, String region,
	  Properties result, String keyName) throws IOException,
	  InterruptedException
  {
	String keypairCmd = EC2_ADD_KEYPAIR_COMMAND + " -K " + pk + " -C " + cert;
	if (region != null)
	  keypairCmd += " --region " + region;
	// Generate a keypair name from the current time for uniqueness
	keypairCmd += " " + keyName;
	NativeCommandExec keypair_exec = new NativeCommandExec();
	int status = keypair_exec.executeNativeCommand(keypairCmd, null, null, 0,
	    false);

	if (keyPairLogger.isInfoEnabled())
	  keyPairLogger.info("Creating keypair " + keyName);

	debugCommandOutput(keypairCmd, status, keypair_exec);

	if (status == 0)
	{ // Add the key pair to the result
	  String keyPair = "";
	  boolean firstLine = true;
	  for (String line : keypair_exec.getStdout())
	  {
		if (firstLine) // Ignore first line
		  firstLine = false;
		else
		  keyPair += line + "\n";
	  }
	  result.setProperty(AmazonProperties.EC2_KEY_PAIR, keyPair);
	  keyPairLogger.info(keyPair);
	}
	else
	{ // Command failed
	  prepareResultError(EC2_ADD_KEYPAIR_COMMAND, keypair_exec, result);
	}
  }

  /**
   * Delete a key pair
   * 
   * @param pk
   * @param cert
   * @param region
   * @param keyName
   * @param result
   * @throws IOException
   * @throws InterruptedException
   */
  private void deleteKeyPair(String pk, String cert, String region,
	  String keyName, Properties result) throws IOException,
	  InterruptedException
  {
	int status;
	String keypairCmd = EC2_DELETE_KEYPAIR_COMMAND + " -K " + pk + " -C "
	    + cert;
	if (region != null)
	  keypairCmd += " --region " + region;
	// Generate a keypair name from the current time for uniqueness
	keypairCmd += " " + keyName;
	NativeCommandExec keypair_exec = new NativeCommandExec();
	status = keypair_exec
	    .executeNativeCommand(keypairCmd, null, null, 0, false);

	if (keyPairLogger.isInfoEnabled())
	  keyPairLogger.info("Deleting keypair " + keyName);
	debugCommandOutput(keypairCmd, status, keypair_exec);

	if (status != 0)
	{ // Command failed
	  prepareResultError(EC2_DELETE_KEYPAIR_COMMAND, keypair_exec, result);
	}
  }

  /**
   * Completion of the commands and return of the results
   * 
   * @param token
   * @param p
   * @param result
   */
  private void returnResult(Object token, Properties p, Properties result)
  {
	// pass the configuration sequence number back
	PropList.setProperty(result, Config.PropertyActionSequenceNumber, p
	    .getProperty(Config.PropertyActionSequenceNumber));
	plugin.configurationComplete(token, result);
  }

  /**
   * Add the proper error code and error message to be returned
   * 
   * @param commandName
   *          name of the command executed
   * @param exec
   *          execution object
   * @param result
   *          properties that are returned
   */
  private void prepareResultError(String command, NativeCommandExec exec,
	  Properties result)
  {
	result.setProperty(Config.PropertyTargetResultCode, Integer
	    .toString(ResultCodeException));
	result.setProperty(PropertyExceptionMessage, command + " execution failed");
	result.setProperty(PropertyExceptionStack, exec.getStderr().toString());
  }

  /**
   * Debug the output of a command being executed
   * 
   * @param commandName
   *          name of the command executed
   * @param status
   *          status returned by the command execution
   * @param exec
   *          execution object
   */
  private void debugCommandOutput(String commandName, int status,
	  NativeCommandExec exec)
  {
	if (logger.isDebugEnabled())
	{
	  logger.debug(commandName + " status: " + status);
	  logger.debug("stdout:" + exec.getStdout());
	  logger.debug("stderr:" + exec.getStderr());
	}
  }

}
