/**
 * DiCloud - GENI project
 * 
 * (C) University of Massachusetts Amherst
 * 
 * Original contributor: Emmanuel Cecchet
 */
package edu.umass.cs.geni.ec2.handlers;

import java.rmi.Naming;
import java.util.MissingResourceException;
import java.util.Properties;

import orca.shirako.plugins.config.Config;
import orca.util.PropList;
import edu.umass.cs.geni.ec2.AmazonProperties;
import edu.umass.cs.geni.ec2.NativeCommandExec;
import edu.umass.cs.geni.ec2.accounting.AWSActivity;
import edu.umass.cs.geni.ec2.accounting.AWSActivityPricing;
import edu.umass.cs.geni.ec2.accounting.S3Proxy;

/**
 * This class defines an S3 handler to manipulate Amazon Simple Storage buckets.
 * 
 * @author <a href="mailto:cecchet@cs.umass.edu">Emmanuel Cecchet </a>
 * @version 1.0
 * 
 */
public class S3Handler extends Config
{
  private static final String S3_CREATE_BUCKET_COMMAND = "s3cmd mb";
  private static final String S3_DELETE_BUCKET_COMMAND = "s3cmd -f rb";
  private static final String S3_PROBE_BUCKET_COMMAND  = "s3cmd info";
  private AWSPlugin           awsPlugin;

  /**
   * Constructor just invoking super
   * 
   * @param plugin
   *          a Shirako plugin
   * 
   */
  public S3Handler(AWSPlugin plugin)
  {
	super(plugin);
	this.awsPlugin = plugin;
  }

  /**
   * Create a new S3 bucket invoking 's3cmd mb'.
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.S3_BUCKET_NAME: name of the S3 bucket to create
   * Optional properties: 
   *  - AmazonProperties.S3_LOCATION: region where to create the bucket
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: command output (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * </pre>
   */
  @Override
  public void join(Object token, Properties p) throws Exception
  {
	if (!awsPlugin.getAwsAccouting().hasCreditLeft())
	  throw new Exception("No AWS credit left to allocate resources");

	// Prepare the command to be executed
	String bucketName = p.getProperty(AmazonProperties.S3_BUCKET_NAME);
	String region = p.getProperty(AmazonProperties.S3_LOCATION);

	// Sanity checks
	if (bucketName == null)
	  throw new MissingResourceException("Missing S3 bucket name in join", this
		  .getClass().getName(), AmazonProperties.S3_BUCKET_NAME);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, TargetJoin);

	// Now create the bucket
	if (logger.isInfoEnabled())
	  logger.info("Creating S3 bucket " + bucketName);

	String command = S3_CREATE_BUCKET_COMMAND + " ";
	if (region != null)
	  command += "--bucket-location=" + region + " ";
	command += "s3://" + bucketName;

	// Let's execute the command
	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	if (status == 0)
	{ // Success, check for information about the instance
	  parseOutputAndSetProperties(result, exec);

	  // Log the create operation that is equivalent to a PUT
	  awsPlugin.getAwsAccouting().logActivity(
		  new AWSActivity("S3", command, AWSActivityPricing.S3_PUT_OPERATION,
		      System.currentTimeMillis()));

	  if (logger.isInfoEnabled())
		logger.info("Registering S3 Proxy for " + bucketName);
	  S3Proxy s3Proxy = new S3Proxy(awsPlugin, p);
	  Naming.rebind(bucketName, s3Proxy);
	}
	else
	{ // Command failed
	  prepareResultError(command, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Delete an S3 bucket by invoking 's3cmd rb'
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.S3_BUCKET_NAME: name of the S3 bucket to create
   * Optional properties: 
   *  - AmazonProperties.S3_LOCATION: region where to create the bucket
   * </pre>
   * 
   */
  @Override
  public void leave(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String bucketName = p.getProperty(AmazonProperties.S3_BUCKET_NAME);
	String region = p.getProperty(AmazonProperties.S3_LOCATION);

	// Sanity checks
	if (bucketName == null)
	  throw new MissingResourceException("Missing S3 bucket name in leave",
		  this.getClass().getName(), AmazonProperties.S3_BUCKET_NAME);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, TargetLeave);

	// Now create the instances
	if (logger.isInfoEnabled())
	  logger.info("Deleting S3 bucket " + bucketName);

	String command = S3_DELETE_BUCKET_COMMAND + " ";
	if (region != null)
	  command += "--bucket-location=" + region + " ";
	command += "s3://" + bucketName;

	// Let's execute the command
	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	/*
	 * The remove force command will generate a warning on stderr if the bucket
	 * is not empty. But NativeCommandExec will return a non 0 status if there
	 * is something on stderr.
	 */
	if ((status == 0) || (exec.getStderr().toString().contains("WARNING")))
	{ // Success
	  parseOutputAndSetProperties(result, exec);

	  if (logger.isInfoEnabled())
		logger.info("De-registering S3 Proxy for " + bucketName);
	  Naming.unbind(bucketName);
	}
	else
	{ // Command failed
	  prepareResultError(command, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Invoke 's3cmd info' to probe an S3 bucket. The expected properties are:
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.S3_BUCKET_NAME: name of the S3 bucket to create
   * Optional properties: 
   *  - AmazonProperties.S3_LOCATION: region where to create the bucket
   * </pre>
   * 
   */
  public void probe(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String bucketName = p.getProperty(AmazonProperties.S3_BUCKET_NAME);
	String region = p.getProperty(AmazonProperties.S3_LOCATION);

	// Sanity checks
	if (bucketName == null)
	  throw new MissingResourceException("Missing S3 bucket name in probe",
		  this.getClass().getName(), AmazonProperties.S3_BUCKET_NAME);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, AWSPlugin.TargetProbe);

	// Now create the instances
	if (logger.isInfoEnabled())
	  logger.info("Probing S3 bucket " + bucketName);

	String command = S3_PROBE_BUCKET_COMMAND + " ";
	if (region != null)
	  command += "--bucket-location=" + region + " ";
	command += "s3://" + bucketName;

	// Let's execute the command
	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	if (status == 0)
	{ // Success, check for information about the instance
	  parseOutputAndSetProperties(result, exec);
	}
	else
	{ // Command failed
	  prepareResultError(command, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Parse the output of the S3 create or describe commands
   * 
   * @param result
   * @param exec
   */
  private void parseOutputAndSetProperties(Properties result,
	  NativeCommandExec exec)
  {
	result.setProperty(Config.PropertyTargetResultCode, Integer
	    .toString(ResultCodeOK));
	result.setProperty(Config.PropertyTargetResultCodeMessage, exec.getStdout()
	    .toString());
  }

  /**
   * Completion of the commands and return of the results
   * 
   * @param token
   * @param p
   * @param result
   */
  private void returnResult(Object token, Properties p, Properties result)
  {
	// pass the configuration sequence number back
	PropList.setProperty(result, Config.PropertyActionSequenceNumber, p
	    .getProperty(Config.PropertyActionSequenceNumber));
	plugin.configurationComplete(token, result);
  }

  /**
   * Add the proper error code and error message to be returned
   * 
   * @param commandName
   *          name of the command executed
   * @param exec
   *          execution object
   * @param result
   *          properties that are returned
   */
  private void prepareResultError(String command, NativeCommandExec exec,
	  Properties result)
  {
	result.setProperty(Config.PropertyTargetResultCode, Integer
	    .toString(ResultCodeException));
	result.setProperty(PropertyExceptionMessage, command + " execution failed");
	result.setProperty(PropertyExceptionStack, exec.getStderr().toString());
  }

  /**
   * Debug the output of a command being executed
   * 
   * @param commandName
   *          name of the command executed
   * @param status
   *          status returned by the command execution
   * @param exec
   *          execution object
   */
  private void debugCommandOutput(String commandName, int status,
	  NativeCommandExec exec)
  {
	if (logger.isDebugEnabled())
	{
	  logger.debug(commandName + " status: " + status);
	  logger.debug("stdout:" + exec.getStdout());
	  logger.debug("stderr:" + exec.getStderr());
	}
	if (status != 0)
	  logger.warn("Command " + commandName + " failed with error: "
		  + exec.getStderr());
  }

}
